/* *********************************************************************************************************************
KVB  C(K)ontrôle de vitesse par balises JLF : 15/02/2025   Version 1.2

V1.0 : 18/12/2024 : Version finalisée.
V1.1 : 06/01/2025 : Version avec trame des roues codeuses entre crochets = <1234561> + Alarme sonore en cas d'appui sur le bouton [TEST] + Séquence d'initialisation.
V1.2 : 15/02/2025 : Version aussi compatible avec BB7200 SimExpress. Ajout 'n'="PA400", 'N'="UC512"+(SOL)+(ENG), 'o'=(SOL)+(ENG) et 'O'="   " après 'N'.

!!! Mettre tous les 'Serial.print()' en commentaire dans la version finale !!!

  Animation des afficheurs et des voyants d'un boitier KVB d'origine.
  Reçoit des ordres par la liaison série RX/TX et envoi des octets en fonction des boutons poussoirs enfocncés.
  Gére aussi un boitier optionnel de roues codeuses et ses haut-parleur et buzzer.
  Ce code peut aussi servir à fabriquer un nouveau boitier KVB.

  La vitesse de la liaison série est de 9600 bps, et peut être facilement modifiée dans le code.

  A la mise sous tension, si le module TM1638 n'est pas détecté, on ne peut rien afficher. Dans ce cas la lampe LSSF clignote rapidement.
  A la mise sous tension, le KVB affiche une intro, puis les 6 chiffres des roues codeuses si le boitier des roues codeuses est présent.
  L'appui de la touche [VAL] pendant plus de 3 secondes, permet de modifier la luminosité des afficheurs et des voyants.

  Site web : http://www.la-tour.info/uts/uts_page15.html

ATTENTION : J'ai utilisé la librairie TM1638 de Ricardo Batista : https://github.com/rjbatista/tm1638-library
            Je l'ai modifié pour pouvoir allumer directement les afficheurs et les voyants du KVB.
            Les programmes : TM1638QYF_JLF.h, TM1638QYF_JLF.cpp, TM163XXFonts_JLF.h, TM18XX_JLF.h et TM16XX_JLF.cpp
            sont modifiés. ne pas les mettre à jour.
            Cette librairie est l'une des rares prévues pour des afficheurs à Anode commune, comme celle du KVB original.

******************
Ce logiciel est un logiciel libre. Exigence du concepteur : Ne pas modifier les lignes :
A la mise sous tension, affiche "JLF xx/xx/xxxx" sur écran pendant 1 secondes.
On peut modifier ce programme et le diffuser. Dans ce cas, il faut préciser l'origine et donner accès aux sources modifiées.
Domaine : Informatique. Définition : Un logiciel libre est un logiciel distribué avec l’intégralité de ses programmes-sources, afin que l’ensemble des utilisateurs qui l’emploient,
puissent l’enrichir et le redistribuer à leur tour.
Note : Un logiciel libre n’est pas nécessairement gratuit et les droits de la chaîne des auteurs sont préservés.
Équivalent étranger : free software, open source software.
(Source : Vocabulaire de l’informatique (liste de termes, expressions et définitions adoptés), NOR: CTNX0710138K, J.O n° 93 du 20 avril 2007 page 7078, texte n° 84)
logiciel libre. Par logiciel libre on entend un logiciel qui offre la liberté aux utilisateurs d'exécuter, de copier, de distribuer, d'étudier, de modifier et d'améliorer le logiciel.
Plus précisément, elle fait référence à quatre types de liberté pour les utilisateurs du logiciel :
    La liberté d'exécuter le programme, pour tous les usages (liberté 0).
    La liberté d'étudier comment le programme fonctionne et de l'adapter à ses besoins (liberté 1). L'accès au code source est une condition requise.
    La liberté de redistribuer des copies, (liberté 2).
    La liberté d'améliorer le programme et de diffuser les améliorations au public pour en faire profiter toute la communauté (liberté 3). L'accès au code source est une condition requise.
***********************

Fonctionne sur un ARDUINO NANO + une carte TM1638 + deux cartes MCP23017 optionnels si on utilise un boitier de roues codeuses.
  Afficheurs : [888][888]     (V) ( FU)      (LSSF)
  Lampes :     [VAL] [MV] [FC] [    ]   [  ]
  Boutons :    [VAL] [MV] [FC] [TEST]   [SF]
                              (SOL) (ENG)
  Roues Cod. : [VD VU]    [VU]    |DC DD DU]    Centaines, Dizaines, Unités + Bouton [VO-ME-MA]

Le programme simule le fonctionnement interne du boitier KVB.
Pictogramme à afficher en fonction du caractère reçu sur la ligne série RX0.
L'afficheur :
 r => [JLF][   ] Reset logicielle Arduino
 R => [   ][   ] Raz éteint tous les afficheurs, voyants et touches
 b => [   ][ b ]
 c => [   ][   ] Efface que l'afficheur [   ][   ]
 e => [   ][===]
 F => [   ][ F ]
 l => [ L ][   ]
 L => [   ][ L ]
 m => [---][---] Moins
 M => [---][---] Moins Clignotant
 p => [ p ][   ]
 P => [   ][ P ]
 t => [888][888] Affiche 888 888
 U => [   ][ FU] fU
 y => [ 00][   ] 2 Zéro
 Y => [000][   ] 3 Zéro
 x => [---][ 00] Moins + 2 Zéro
 X => [ 00][000] 2 Zéro + 3 Zéro
 z => [   ][ 00] 2 Zéro Clignotant
 Z => [   ][ 00] 2 Zéro Fixe
 n => [ PA][400] Affichage = PA400 + Buzzer. Spécial pour BB7200 de SimExpress
 N => { UC][512] + Lampes (SOL) et (ENGIN) allumées + HP. Spécial pour SimExpress
 o =>              Lampes (SOL) et (ENGIN) allumées + Buzzer + HP. Spécial SimExpress
 O => [   ][   ] + Lampes (SOL) et (ENGIN) éteintes + Stop Buz+HP. Spécial SimExpress
 s => Séquence d'initialisation du KVB, à la mise sous tension du pupitre.
 ? => Demande si le KVB est pret ? Réponse = 'o' pour ok, 'n' pour nok si pas de carte TM1638 détectée.
 ! => Demande le n° des roues codeuses + bouton VO-ME-MA.  Exemple '<1234561>' ou '<0000000>' si pas de boitier de RC.
      Attention, ce n'est pas le nombre validé par le bouton [VAL], mais la position actuelle des rc. Le dernier chiffre = 1:VO, 2:ME, 3:MA.
 v => Demande la version du KVB. En retour le boitier envoie : JLF_Vx.x_jj/mm/aaaa

Les voyants :
 0 =>  (V)   Eteint     1 =>  (V)   Allumée
 2 =>  (FU)  Eteint     3 =>  (FU)  Allumée
 4 => (SOL)  Eteint     5 => (SOL)  Allumée
 6 => (ENG)  Eteint     7 => (ENG)  Allumée
 8 => Eteint toutes les lampes et LSSF (Mais pas les touches)
 g => [VAL]  Eteint     G => [VAL]  Allumée     A => [VAL]  Clignotant
 h =>  [MV]  Eteint     H =>  [MV]  Allumée
 i =>  [FC]  Eteint     I =>  [FC]  Allumée
 j => (LSSF) Eteint     J => (LSSF) Fixe        C => (LSSF) Clignotant
 9 => Eteint toutes les touches (Mais pas les voyants ni LSSF)
 k => Arret klaxon      K => Klaxon actif sur haut-parleur
 q => Arret buzzer      Q => Buzzer actif
 T => Fonction Test complet pendant 3 secondes

Le code des touches envoyées sur la liaison série :
On envoie un caractère par touche.
   [VAL] => 'A' enfoncée, 'a' relâchée
    [MV] => 'B' enfoncée, 'b' relâchée
    [FC] => 'C' enfoncée, 'c' relâchée
  [TEST] => 'D' enfoncée, 'd' relâchée
    [SF] => 'E' enfoncée, 'e' relâchée
KVB pret => 'o'   ('o' pour ok 'n'  pour nok, si la carte TM1638 n'est pas détectée).

Branchement :
               ARDUINO NANO
   SIMU--<--TX1            VIN <-- 12 Volts
   SIMU-->--RX0            GND
            RESET        RESET
            GND            +5V
            D02            A07
            D03            A06
            D04            A05---I2C-SCL
            D05            A04---I2C-SDA
 TM   DIO---D06            A03
 16   CLK---D07            A02
 38   STK---D08            A01
      LSSF--D09            A00
       HP---D10           VREF
            D11           3,3V
            D12            D13

L'utilisation de la sortie HP est optionnelle. On l'utilisera si le son issu du MCP23017 n'est pas satisfaisant.


Pour la lecture des roues codeuses, on écrit dans le MCP23017_SORTIES un bit à '1' de 0 à '9'.
Ensuite, on lit dans le second MC23017_ENTREES un octet pour connaitre le chiffre de la roue codeuse.
              0             1             2             3             4             5             6
           Roue V        Roue V        Roue L        Roue D        Roue D        Roue D        Bouton
           Dizaine       Unité         Unité        Centaine       Dizaine       Unité        VO ME MA
              |             |             |             |             |             |             |
  0 ----------+-------------+-------------+-------------+-------------+-------------+             |
  1 ----------+-------------+-------------+-------------+-------------+-------------+-------------+
  2 ----------+-------------+-------------+-------------+-------------+-------------+-------------+
  3 ----------+-------------+-------------+-------------+-------------+-------------+-------------+
  4 ----------+-------------+-------------+-------------+-------------+-------------+
  5 ----------+-------------+-------------+-------------+-------------+-------------+
  6 ----------+-------------+-------------+-------------+-------------+-------------+
  7 ----------+-------------+-------------+-------------+-------------+-------------+
  8 ----------+-------------+-------------+-------------+-------------+-------------+
  9 ----------+-------------+-------------+-------------+-------------+-------------+
 10
 11
 12
 13  --->   HP    Sortie signal continu 0/5 Volts : 5 Volts = En marche.
 14  ---> Buzzer  Sortie signal continu 0/5 Volts : 0 Volt  = En marche.
 15  --->   HP    Sortie modulée Fréquence 1 Khz.

                              MCP23017_ENTREES  Adresse = 0x21                                                 MCP23017_SORTIES  Adresse = 0x20
      +---+---+-------+-------+-------+-------+-------+-------+---------+-----+      +---+---+----+----+----+----+----+-----+----+-----+
      | A2| A1| A0 +5V|       |       |       |       |       |   GND   | +5V |      | A2| A1| A0 |    |    |    |    |     |GND | +5V |
      +---+---+-------+-------+-------+-------+-------+-------+---------+-----+      +---+---+----+----+----+----+----+-----+----+-----+

      +---+---+-------+-------+-------+-------+-------+-------+---------+-----+      +---+---+----+----+----+----+----+-----+----+-----+
      |+5V| X |   B0  |   B1  |   B2  |   B3  |   B4  |   B5  |   B6    |  B7 |      |+5V| X | B0 | B1 | B2 | B3 | B4 | B5  | B6 | B7  |
      |   |   |       |       |       |       |       |       |         |     |      |   |   |  8 |  9 |    |    |    |HPfix|BUZZ|HPmod|
      +---+---+-------+-------+-------+-------+-------+-------+---------+-----+      +---+---+----+----+----+----+----+-----+----+-----+
      |GND| X |   A0  |   A1  |   A2  |   A3  |   A4  |   A5  |   A6    |  A7 |      |GND| X | A0 | A1 | A2 | A3 | A4 | A5  | A6 | A7  |
      |   |   |  V_D  |  V_U  |  L_U  |  D_C  |  D_D  |  D_U  |bt_vomema|     |      |   |   |  0 |  1 |  2 |  3 |  4 |  5  |  6 |  7  |
      +---+---+-------+-------+-------+-------+-------+-------+---------+-----+      +---+---+----+----+----+----+----+-----+----+-----+

HPmod = Sortie modulée, on peut y brancher directement le HP avec un résistance de 150 Ohms en série.
HPfix = Sortie continue 0 ou 5 Volts. Il faut ajouter un circuit de type sirène pour y brancher le HP.

*/
// ------------------------------------------------------------------------------------------------------------------------------------
#include <EEPROM.h>                       // Pour enregistrer la valeur de la luminosité.
#include <avr/wdt.h>                      // Pour le reset logiciel par la fonction : wdt_enable(prescaller).
#include "Wire.h"                         // Pour les broches I2C.
#include "TM1638QYF_JLF.h"                // Pour le modude TM1638. Librairie modifiée de Ricardo Batista  URL : https://github.com/rjbatista/tm1638-library
#include "MCP23017.h"                     // Pour les modules I2C de lecture roues codeuses. Librairie de Rob Tillaart (V0.8.0)  URL: https://github.com/RobTillaart/MCP23017_RT

//      Plot RX0                          // Plots de liaison série.
//      Plot TX1                          //
//      Plot I2C_SDA   4                  // Plots de communications I2C avec les MCP23017, déclaré dans 'Wire.h'.
//      Plot I2C_SCL   5                  // ..
#define SORTIE_TM_DIO  6                  // Plots de communications propriétaires avec le TM1638.
#define SORTIE_TM_CLK  7                  // ..
#define SORTIE_TM_STB  8                  // ..
#define SORTIE_LSSF    9                  // Sortie. Signal de commande lampe LSSF : 1 = Allumée.
#define SORTIE_HP     10                  // Sortie. Signal BF 'pwm' pour alarme sonore, pour aller vers un amplificateur optionnel.

// Fonctions désactivables par l'utilisateur, si le #define est mis en commentaires = //#define ....
// On choisi si c'est le boitier KVB qui assure ces fonctions en interne, ou si c'est réalisé par l'échange sur la liaison série.
#define FONCTION_TEST                     // Permet de lancer la fonction test avec l'appui de la touche [TEST].
#define FONCTION_AFF_ROUES                // Permet d'afficher le n° des roues codeuses au démarrage, pendant 1 seconde.
#define FONCTION_VAL_ROUES                // Permet de faire clignoter le bouton [VAL] si l'on modifie les roues codeuses.
#define FONCTION_ENV_ROUES                // Permet d'envoyer les 7 chiffres des roues codeuses sur la liaison série, une fois validé par le bouton [VAL].
#define FONCTION_OK_MARCHE                // Envoi un 'o' ou un 'n' sur la liaison série, une fois le KVB en ordre de marche.
#define FONCTION_VOY_MV_FC                // Allume les touches [MV] et [FC] en cas d'appui dessus.
#define FONCTION_SF_LSSF                  // Un appui sur le bouton [SF] éteint le voyant LSSF.
#define FONCTION_VAL_VAL                  // Un appui sur le bouton [VAL] éteint le bouton [VAL].

// Constantes modifiables par l'utilisateur.
const long duree_antirebond = 300;        // Durée anti rebond des boutons = 300 msec par défaut.
const long duree_timer_clignotant = 300;  // Durée clignotant du LSSF et de [000] = 300 msec par défaut.
const long duree_timer_test_local = 3000; // Durée du test local du KVB = 3 sec par défaut.
const long duree_timer_luminosite = 3000; // Temps d'appui sur la touche [VAL] pour passer en mode réglage luminosité = 3 sec par défaut.

char buffer[20], buffer_old[20], buffer_cli[20];   // Chaine de caractères envoyée au TM1638.
byte lampe_V, lampe_FU, lampe_SOL;                 // Lampes des touches:  0 = Eteint, 1 = Allumé.
byte lampe_ENG, voyant_VAL, voyant_VALc;           // Voyants:       0 = Eteint,  1 = Allumé.
byte voyant_MV, voyant_FC;                         // Voyants:       0 = Eteint,  1 = Allumé.
byte ampoule_LSSF;                                 // Voyant LSSF:   0 = Eteint,  1 = Allumé, 2 = Clignotant.
byte klaxon;                                       // Klaxon sur HP: 0 = Eteint,  1 = Actif.
byte buzzer;                                       // Buzzer:        0 = Eteint,  1 = Actif.
byte bouton_VAL, bouton_MV, bouton_FC;             // Touches:       0 = Relâché, 1 = Enfoncé.
byte bouton_TEST, bouton_SF;                       // Touches:       0 = Relâché, 1 = Enfoncé.
byte bouton_VAL_old, bouton_MV_old, bouton_FC_old; // Touches: Ancien état.
byte bouton_TEST_old, bouton_SF_old;               // Touches: Ancien état.
unsigned long bouton_VAL_milli, bouton_MV_milli;   // Timer d'antirebond des touches.
unsigned long bouton_FC_milli, bouton_TEST_milli;  // Timer d'antirebond des touches.
unsigned long bouton_SF_milli, courante_milli;     // Timer d'antirebond des touches.
unsigned long timer_clignotant;                    // Timer fréquence du clignotement.
unsigned long timer_test_local;                    // Timer durée du test local (Bouton [TEST]).
unsigned long timer_luminosite;                    // Timer sur la touche [VAL] déclenchant le réglage de luminosité.
unsigned long timer_sequence_init;                 // Timer dans la séquence d'intitialisation.
byte tm1638_ok;                      // 0/1 : 1 = Le TM1638 est présent.
byte sequence_init_num;              // Numéro de la séquence d'initialisation en cours.
byte maj;                            // 0/1 : 1 = Mise à jour de l'afficheur, lampes ou voyants demandé !
byte i;
byte etat_boutons;                   // En retour de lecture du clavier du TM1638.
char car_recu;                       // Caractère reçu sur la ligne série RX0.
byte aff_clignotant, var_clignotant; // Afficheur de droite: 0 = Fixe ou Eteint. 1 = Clignotant.
byte test_local;                     // 0/1 : 1 = Test local du KVB en cours.
byte luminosite;                     // 0à7 : Luminosité de l'affichage.
byte test_luminosite;                // 0/1 : 1 = Réglage luminosité en cours.

unsigned int  roue_1, roue_2;        // Pour la lecture des roues codeuses.
unsigned int  roue_cpt1;
byte roue_cpt2;
byte sortie10, sortie11, sortie12, sortie13, sortie14, sortie15;
byte mcp23017_ok;                    // 0/1 : 1 = Les deux MCP23017 sont présents.
byte mcp23017_mode;
byte roue_V_D, roue_V_U, roue_L_U;   // Valeur des roues codeuses, VV L DDD, Centaines, Dizaines, Unités.
byte roue_D_C, roue_D_D, roue_D_U;   // 0 à 9.
byte bt_vomema;                      // 1 = VO, 2 = ME, 3 = MA.
byte roue_cheksum;                   // Permet de savoir si une roue codeuse a bougé.
byte roue_valid;                     // 0/1 : 1 = Si une roue codeuse a bougé, on doit valider en appuyant sur [VAL].

TM1638QYF afficheurTM(SORTIE_TM_DIO, SORTIE_TM_CLK, SORTIE_TM_STB); // Déclaration du module TM1638 sur 3 fils : DIO, CLK, STB.
MCP23017  MCP_SORTIES(0x20);                                        // Module MCP23017 pour les 16 Sorties. Adresse I2C = 0x20.
MCP23017  MCP_ENTREES(0x21);                                        // Module MCP23017 pour les 16 Entrées. Adresse I2C = 0x21.

// ------------------------------------------------------------------------------------------------------------------------------------
// ------------------------------------------------------------------------------------------------------------------------------------
// ------------------------------------------------------------------------------------------------------------------------------------
void setup() {
  pinMode(SORTIE_LSSF, OUTPUT);      // Plot de sortie: Commande du voyant LSSF.
  pinMode(SORTIE_HP, OUTPUT);        // Plot de sortie: La broche SORTIE_HP vers un ampli est en sortie.
  analogWrite(SORTIE_HP, 0);         // Génération du signal pwm pour l'alarme sonore. 0 = A l'arrêt.
  courante_milli = millis();         // Initialise la variable 'courante_milli'.
  delay(100);                        // Attendre 100 milli secondes.
  Wire.begin();                      // Affectation des broches SDA + SCL.
  luminosite = EEPROM.read(0);       // Lecture de l'epprom pour récuperer la valeur de la luminosité = 'luminosite'.
  if (luminosite > 7) { luminosite = 7 ; EEPROM.update(0, 7); }
  afficheurTM.setupDisplay(true, luminosite); // Activer l'affichage du TM1638 avec une intensité de 0 à 7.
  delay(100);                        // Attendre 100 milli secondes.
  etat_boutons = afficheurTM.getButtons();

// Pour savoir si le TM1638 est branché, on lit la position des boutons poussoirs. Normalement, ils sont à '0', sauf si le TM1638 n'est pas branché.
// Dans ce cas on lit 'FF' pour les 8 boutons. On teste les boutons 6, 7 et 8 car ils ne sont pas branchés, donc on ne risque pas d'appuyer dessus.
// Si la lecture des boutons retourne '111xxxxx', le TM1638 n'est pas connecté, ou mal connecté (DIO, CLK, STB).
  if ((etat_boutons & 0xE0) == 0xE0) tm1638_ok = 0 ; else tm1638_ok = 1;  // 1 = Le TM1638 est présent. On suppose les boutons 6, 7 et 8 non appuyés.
  Serial.begin(9600);                // Ouverture de la liaison série RT/TX.
  //Serial.print( etat_boutons);  Serial.print(' '); Serial.print(tm1638_ok);

// -----------------------------------------------
// Initialisation des variables.
  lampe_V = 0;        lampe_FU = 0;      lampe_SOL = 0;     lampe_ENG = 0;       ampoule_LSSF = 0;
  voyant_VAL = 0;     voyant_VALc = 0;   voyant_MV = 0;     voyant_FC = 0;
  bouton_VAL = 0;     bouton_MV = 0;     bouton_FC = 0;     bouton_TEST = 0;     bouton_SF = 0;
  bouton_VAL_old = 0; bouton_MV_old = 0; bouton_FC_old = 0; bouton_TEST_old = 0; bouton_SF_old = 0;

  buffer[0] = '\0';   buffer_cli[0] = '\0'; buffer_old[0] = '\0'; // Il faut que les chaines de caractères soient initialisées.

  bouton_VAL_milli  = courante_milli;  // Initialise les durées des antirebonds. La variable 'courante_milli' est déjà initialisé.
  bouton_MV_milli   = courante_milli;
  bouton_FC_milli   = courante_milli;
  bouton_TEST_milli = courante_milli;
  bouton_SF_milli   = courante_milli;
  timer_clignotant  = courante_milli;
  timer_test_local  = courante_milli;

  aff_clignotant    = 0;
  var_clignotant    = 0;
  test_local        = 0;  // Pas d'affichage clignotant au début.
  maj               = 0;
  etat_boutons      = afficheurTM.getButtons();  // etat_boutons => 1 bit "0/1" par bouton.
  test_luminosite   = 0;
  if (tm1638_ok == 0) ampoule_LSSF = 2; // Si TM1638 absent fait clignoter le vooyant LSSF

// -----------------------------------------------
// Partie pour le boitier optionnel des roues codeuses.
// Si le boitier des roues codeuses est connecté alors : mcp23017_ok = 1.
  mcp23017_ok = 1;
  if (MCP_SORTIES.begin() == 0) { mcp23017_ok = 0; }   // Module MCP23017 pour les 16 sorties absent.
  else {
    MCP_SORTIES.write16(uint16_t(0x00C0));        // Sorties 13 = 0, 14 et 15 = 1 pour ne pas avoir d'alarme sonore. Attention les 8 bits forts sont cadrés à droite.
    MCP_SORTIES.pinMode16(uint16_t(0x0000));      // Module MCP configuré en 16 sorties. (0 = output , 1 = input).
    MCP_SORTIES.setPullup16(uint16_t(0x0000));    // Module MCP configuré sans les résistances de pull-up.
    MCP_SORTIES.setPolarity16(uint16_t(0x0000));  // Module MCP configuré sans inversion de polarité en entrée.
  }
  if (MCP_ENTREES.begin() == 0) { mcp23017_ok = 0; }   // Module MCP23017 pour les 16 entrées absent.
  else {
    MCP_ENTREES.pinMode16(uint16_t(0x7F7F));      // Module MCP configuré en 14 entrées. (0 = output , 1 = input).
    MCP_ENTREES.setPullup16(uint16_t(0xFFFF));    // Module MCP configuré avec les résistances de pull-up.
    MCP_ENTREES.setPolarity16(uint16_t(0x0000));  // Module MCP configuré sans inversion de polarité en entrée.
  }
  //Serial.print(" mcpB");Serial.print(mcp23017_ok);Serial.print(" ");
  roue_V_D = 0; roue_V_U = 0; roue_L_U = 0; roue_D_C = 0; roue_D_D = 0; roue_D_U = 0; bt_vomema = 0;
  mcp23017_mode = 0;
  roue_cpt1 = 2;         // Ne pas modifier cette valeur : roue_cpt1 = 2;
  roue_cpt2 = 0;         // Ne pas modifier cette valeur : roue_cpt2 = 0; roue_cpt1 et roue_cpt2 vont ensemble !
  roue_1   = 0;
  sortie10 = 0; sortie11 = 0; sortie12 = 0;
  sortie13 = 0; sortie14 = 0; sortie15 = 0;
  klaxon = 0; buzzer = 0;

// -----------------------------------------------
  intro_jlf();                         // Affichage sur l'écran d'une intro sympa signée 'JLF'.
  strncpy(buffer,     "        ", 8);  // 8 Espaces = Affichage éteint.
  strncpy(buffer_old, "        ", 8);  // 8 Espaces = Affichage éteint.
  strncpy(buffer_cli, "        ", 8);  // 8 Espaces = Affichage éteint.

// -----------------------------------------------
// Si le boitier des roues codeuses est relié au KVB, on affiche pendant 1 sec la valeurs des roues codeuses.
// Il faut passer plusieurs fois dans la routine pour lire toutes les roues codeuses (> 20 fois).
#ifdef FONCTION_AFF_ROUES                        // FONCTION : Permet d'afficher le n° des roues codeuses au démarrage, pendant 1 seconde.
  delay(500);  // Attendre 500 milli secondes.   // Pour ne pas afficher les roues immédiatement.
  if ((mcp23017_ok == 1) && (tm1638_ok == 1)) {
    for (i = 0; i <25 ; i++) lect_roues_cod();   // On appelle la routine de lecture 25 fois.
    buffer[2] = '0'+ roue_V_D; buffer[3] = '0'+ roue_V_U; buffer[4] = '0'+ roue_L_U;
    buffer[5] = '0'+ roue_D_C; buffer[6] = '0'+ roue_D_D; buffer[7] = '0'+ roue_D_U;
    afficheurTM.setDisplayToString(buffer);      // Affichage = N° des roues codeuses pendant 1 seconde.
    delay(1000);
    strncpy(buffer, "        ", 8);
    afficheurTM.setDisplayToString(buffer);      // Raz affichage.
    roue_cheksum = roue_V_D + roue_V_U + roue_L_U + roue_D_C + roue_D_D + roue_D_U + bt_vomema; // Pour savoir si l'on tourne une roue codeuse.
  }
#endif

#ifdef FONCTION_OK_MARCHE                                        // FONCTION : Envoi un 'o' ou un 'n' sur la liaison série, une fois le KVB en ordre de marche.
  if (tm1638_ok == 1) Serial.write('o'); else Serial.write('n'); // Envoie 'o' pour ok, ou 'n' pour nok si TM1638 absent.
#endif
  sequence_init_num = 0;
}

// ------------------------------------------------------------------------------------------------------------------------------------
// ------------------------------------------------------------------------------------------------------------------------------------
// ------------------------------------------------------------------------------------------------------------------------------------
void loop() {
// Réception série sur la ligne RX0 des ordres reçu par le KVB. Un octet = Une commande à executer par le KVB.
  if ((Serial.available() > 0) && (sequence_init_num == 0) && (test_local == 0)) {  // On a reçu un caractère ? et on n'est pas dans la séquence d'initialisation ou de test.
    car_recu = Serial.read();
    switch (car_recu) {
      case '0': lampe_V   = 0; break;     // Voyant  (V)    Eteint / Allumé.
      case '1': lampe_V   = 1; break;     //
      case '2': lampe_FU  = 0; break;     // Voyant  (FU)   Eteint / Allumé.
      case '3': lampe_FU  = 1; break;     //
      case '4': lampe_SOL = 0; break;     // Voyant  (SOL)  Eteint / Allumé.
      case '5': lampe_SOL = 1; break;     //
      case '6': lampe_ENG = 0; break;     // Voyant (ENGIN) Eteint / Allumé.
      case '7': lampe_ENG = 1; break;     //
      case 'g': voyant_VAL = 0; voyant_VALc = 0; break; // Voyant touche [VAL] Eéteint / Allumé.
      case 'G': voyant_VAL = 1; voyant_VALc = 0; break; //
      case 'A': voyant_VAL = 1; voyant_VALc = 1; break; // Voyant touche [VAL] Clignotant.
      case 'h': voyant_MV = 0; break;     // Voyant touche [MV] Eteint / Allumé
      case 'H': voyant_MV = 1; break;     //
      case 'i': voyant_FC = 0; break;     // Voyant touche [FC] Eteint / Allumé
      case 'I': voyant_FC = 1; break;     //
      case 'j': ampoule_LSSF = 0; break;  // Voyant (LSSF) Eteint / Allumé
      case 'J': ampoule_LSSF = 1; break;  //
      case 'C': ampoule_LSSF = 2; break;  // Voyant (LSSF) Clignotant
      case '8': lampe_V    = 0; lampe_FU     = 0; lampe_SOL = 0; // Eteint les 4 voyants + LSSF, mais pas les touches.
                lampe_ENG  = 0; ampoule_LSSF = 0; break;
      case '9': voyant_VAL = 0; voyant_VALc  = 0; voyant_MV = 0; // Eteint les 3 touches, mais pas les voyants.
                voyant_FC  = 0; break;
      case 'r': softwareReset(WDTO_60MS); break;                            // Fait un 'Reset logiciel'.
      case 'R': strncpy(buffer, "        ", 8); aff_clignotant = 0;         // Eteint tout sur la façade.
                lampe_V   = 0; lampe_FU     = 0; lampe_SOL  = 0;
                lampe_ENG = 0; ampoule_LSSF = 0; voyant_VAL = 0; voyant_VALc = 0;
                voyant_MV = 0; voyant_FC    = 0; klaxon     = 0; buzzer = 0;  analogWrite(SORTIE_HP, 0); break;
      case 's': sequence_init_num = 0; sequence_init(); break;              // Séquence d'initialisation du KVB, à la mise sous tension du pupitre.
      case 'T': test_local = 1; timer_test_local = courante_milli + duree_timer_test_local; // Idem on a appuyé sur la touche [TEST].
                analogWrite(SORTIE_HP, 30);                                 // Génération du signal PWM, pour alarme sonore. 30=ON.
                klaxon = 1; buzzer = 1;                                     // Active le HP et le buzzer via le MCP23017 + buzzer.
                afficheurTM.setDisplayToString("ss888888");                 // Allume tous les voyants et les lampes + affichage = [888][888].
                ampoule_LSSF = 1; break;                                    // Voyant LSSF allumé.
      case 'b': strncpy(buffer, "      b ", 8); aff_clignotant = 0; break;  // 8 caractères.
      case 'c': strncpy(buffer, "        ", 8); aff_clignotant = 0; break;  // 8 caractères.
      case 'e': strncpy(buffer, "     ttt", 8); aff_clignotant = 0; break;  // 8 caractères. Affiche [   ][===]
      case 'F': strncpy(buffer, "      F ", 8); aff_clignotant = 0; break;  // 8 caractères.
      case 'l': strncpy(buffer, "   L    ", 8); aff_clignotant = 0; break;  // 8 caractères.
      case 'L': strncpy(buffer, "      L ", 8); aff_clignotant = 0; break;  // 8 caractères.
      case 'm': strncpy(buffer, "  ------", 8); aff_clignotant = 0; maj = 1; break;  // 8 caractères. Fixe. maj=1, aucas ou l'on affichait des '------' cligontants.
      case 'M': strncpy(buffer, "  ------", 8); aff_clignotant = 1; break;  // 8 caractères. Clignotant.
      case 'p': strncpy(buffer, "   P    ", 8); aff_clignotant = 0; break;  // 8 caractères.
      case 'P': strncpy(buffer, "      P ", 8); aff_clignotant = 0; break;  // 8 caractères.
      case 't': strncpy(buffer, "  888888", 8); aff_clignotant = 0; break;  // 8 caractères.
      case 'U': strncpy(buffer, "      FU", 8); aff_clignotant = 0; break;  // 8 caractères.
      case 'y': strncpy(buffer, "   00   ", 8); aff_clignotant = 0; break;  // 8 caractères.
      case 'Y': strncpy(buffer, "  000   ", 8); aff_clignotant = 0; break;  // 8 caractères.
      case 'x': strncpy(buffer, "  --- 00", 8); aff_clignotant = 0; break;  // 8 caractères.
      case 'X': strncpy(buffer, "   00000", 8); aff_clignotant = 0; break;  // 8 caractères.      
      case 'z': strncpy(buffer, "      00", 8); aff_clignotant = 1; break;  // 8 caractères. Clignotant.
      case 'Z': strncpy(buffer, "      00", 8); aff_clignotant = 0; maj = 1; break;  // 8 caractères. Fixe.  maj=1, aucas ou l'on affichait des '------' cligontants.
      case 'k': klaxon = 0; analogWrite(SORTIE_HP,  0); break;              // Active/Arrête l'alarme sonore sur le haut-parleur.
      case 'K': klaxon = 1; analogWrite(SORTIE_HP, 30); break;              //
      case 'q': buzzer = 0; break;                                          // Active/Arrête l'alarme sonore sur le buzzer.
      case 'Q': buzzer = 1; break;                                          //
      case '?': if ((tm1638_ok == 1) && (Serial.availableForWrite() > 0)) Serial.write('o'); else Serial.write('n');  break; // Demande si le KVB est pret ? Envoie 'o' pour ok, ou 'n' pour nok TM1638 absent.
      case '!': if (Serial.availableForWrite() > 0) {Serial.write('<');Serial.write('0'+roue_V_D); Serial.write('0'+roue_V_U); Serial.write('0'+roue_L_U);
                                                     Serial.write('0'+roue_D_C); Serial.write('0'+roue_D_D); Serial.write('0'+roue_D_U); Serial.write('0'+bt_vomema);Serial.write('>');} break;
      case 'v' :if (Serial.availableForWrite() > 0) {Serial.write("JLF_V1.2_15/02/2025");} // Demande la version du KVB.
      case 'n' :strncpy(buffer, "   PA400", 8); aff_clignotant = 0;                                           buzzer = 1; break; // Affichage = [ PA][400].
      case 'N' :strncpy(buffer, "   UC512", 8); aff_clignotant = 0; lampe_SOL = 1; lampe_ENG = 1; klaxon = 1;             break; // Affichage = [ UC][512] + lampes SOL et ENGIN on.
      case 'o' :                                aff_clignotant = 0; lampe_SOL = 1; lampe_ENG = 1; klaxon = 1; buzzer = 1; break; // Affichage                lampes SOL et ENGIN on.
      case 'O' :strncpy(buffer, "        ", 8); aff_clignotant = 0; lampe_SOL = 0; lampe_ENG = 0; klaxon = 0; buzzer = 0; break; // Affichage = [ UC][512] + lampes SOL et ENGIN off.
      default: break;
    }
  }
// -----------------------------------------------
// Veut-on passer en mode 'réglage de luminosité', en appuyant sur la touche [VAL] pendant 3 secondes ?
  if ((test_luminosite == 1) && (courante_milli > timer_luminosite)) {
    test_luminosite = 2;                                // Oui, on passe en mode 'réglage de luminosité'.
    strncpy(buffer, "  LUN=  ", 8); aff_clignotant = 0; // Afficheur = [LUM][= 5] 5 = Valeur acuelle de la luminosité.
    buffer[7] = '0' + luminosite;                       // La modification de la valeur de la luminosité se fera au niveau du code des touches.
    voyant_VAL = 1; voyant_MV = 1; voyant_FC = 1;       // Voyants des 3 touches, allumés.
  }
  if ( test_luminosite == 2 ){
// On est en mode 'réglage de luminosité', après avoir appuyé sur la touche [VAL] pendant 3 secondes.
    strncpy(buffer, "  LUN=  ", 8); aff_clignotant = 0; // On affichera buffer[] que si la valeur 'luminosite' aura changé.
    voyant_VAL = 1; voyant_MV = 1; voyant_FC = 1;       // Voyants des 3 touches, allumés.
    buffer[7] = '0' + luminosite;                       // La modification de la valeur de la luminosité se fera au niveau du code des touches.
  }
// -----------------------------------------------
  if (test_local == 1) {
// On en mode 'Test local' après avoir appuyé sur la touche [TEST].
    if (courante_milli > timer_test_local) {
      test_local = 0;                                   // On est arrivé à la fin du mode test.
      maj        = 1;                                   // On demande un rafraichissement de l'afficheur.
      analogWrite(SORTIE_HP, 0);                        // Génération du signal PWM, pour alarme sonore. 0=OFF.
      klaxon = 0;                                       // Arrête le klaxon sur haut-parleur.
      buzzer = 0;                                       // Arrête le buzzer.
    }
  }
  else {

// On n'est pas en mode 'Test local'.
// Animation des voyants et des lampes.
// Voir déclaration des lettres 'd' (code 100) à 't' (code 116) dans TM16XXFonts_JLF.h.
//
    buffer[0] = 100 + voyant_VAL + (voyant_MV<<1) + (voyant_FC<<2);                   // Voyants : [VAL](a) + [MV](b) +    [FC](c)
    buffer[1] = 100 + lampe_V    + (lampe_SOL<<1) + (lampe_ENG<<2) + (lampe_FU<<3);   // Lampes  :    V (a) + SOL (b) + ENGINE (c) + FU (e)
    if ((voyant_VALc == 1) && (var_clignotant == 0)) buffer[0] = 100 + (voyant_MV<<1) + (voyant_FC<<2); // Voyants : [VAL] clignotant. ici il est éteint = 0!

// -----------------------------------------------
// Mise à jour de l'affichage [888][888], des 4 lampes et des 3 voyants.
// Uniquement si le buffer[] d'affichage a été modifié, car cela prend du temps l'envoi des donneés au TM1638.
    for (i = 0; i < 8; i++) {
      if (buffer[i] != buffer_old[i]) maj = 1;
    }

// Gestion du clignotement de l'afficheur de droite.
// La variable 'courante_milli' est mise à jour dans la routine = lecture_boutons(). Pas besoin de refaire appel ici à cette routine.
   if (tm1638_ok == 1) {
      if (courante_milli - duree_timer_clignotant > timer_clignotant) {
        timer_clignotant = courante_milli;                   // On a dépassé la durée : 'duree_timer_clignotant'.
        var_clignotant = var_clignotant^1;
        if (aff_clignotant == 1) maj = 1;                    // Au depassement du timer, on force la mise à jour de l'afficheur.
      }
    }
    else { // Si le module TM1638 est absent, la lampe LSSF clignote rapidement.
      if (courante_milli - 100 > timer_clignotant) {
        timer_clignotant = courante_milli;                   // On a dépassé la durée : 100 msec.
        var_clignotant = var_clignotant^1;
        if (aff_clignotant == 1) maj = 1;                    // Au depassement du timer, on force la mise à jour de l'afficheur.
      }
    }

// -----------------------------------------------
// On va envoyer la chaine buffer[] vers l'afficheur.
    if (maj == 1) {                                          // Demande mise à jour de l'affichage du KVB ?
      maj = 0;
      strncpy(buffer_old, buffer, 8);                        // 8 caractères à envoyer.
      if ((aff_clignotant == 1) && (var_clignotant == 0)) {
        strncpy(buffer_cli, buffer, 8);                      // Si l'afficheur doit clignoter,
        buffer_cli[5] = ' ';                                 //
        buffer_cli[6] = ' ';
        buffer_cli[7] = ' ';
        afficheurTM.setDisplayToString(buffer_cli);          // Afficheur de droite clignotant. On affiche la copie de buffer[] sans la partie de droite.
      }
      else afficheurTM.setDisplayToString(buffer);           // Afficheurs fixes.
    }

// -----------------------------------------------
// Le voyant LSSF est traité à part, car c'est une sortie indépendante du reste des voyants.
// C'est relié à une broche de sortie 'SORTIE_LSSF' de l'Arduino.
// Ampoule 'LSSF' : 0 => éteint,  1 => allumée, 2 => clignotant.
    if (ampoule_LSSF == 0) digitalWrite(SORTIE_LSSF, LOW);
    if (ampoule_LSSF == 1) digitalWrite(SORTIE_LSSF, HIGH);
    if (ampoule_LSSF == 2) {
      if (var_clignotant == 1) digitalWrite(SORTIE_LSSF, HIGH); // Sortie clignotante.
      else                     digitalWrite(SORTIE_LSSF, LOW);
    }
  }

// -----------------------------------------------
// Lecture des boutons en face avant du KVB.   [VAL] [MV] [FC] [TEST]   [SF]
// Si une touche est enfoncée ou relâchée, on envoie un caractère.
  lecture_boutons();
  if ((Serial.availableForWrite() > 0) && (sequence_init_num == 0) && (test_local == 0) && (tm1638_ok)) {
// On peut envoyer des octets sur la ligne série. On n'est pas dans la séquence d'initialisation ou de test.
    if (bouton_VAL != bouton_VAL_old) {
// Si l'on appuye sur la touche [VAL] pendant 3 secondes, on passe en mode 'Réglage luminosité'.
      bouton_VAL_old = bouton_VAL;
      if ((bouton_VAL == 0) && (test_luminosite == 1)) test_luminosite = 0; // On a relâché le bouton [VAL] avant de valider le passge en réglage luminosité.
      if (bouton_VAL == 1) Serial.write('A'); else Serial.write('a');       // Envoi le code de la touche [VAL] sur la liaison série.
      if ((bouton_VAL == 1) && (test_luminosite == 0)) {                    // On arme le timer pour passer en mode 'Réglage luminosité' au bout du timer.
        test_luminosite = 1;
        timer_luminosite = courante_milli + duree_timer_luminosite;
       }
      if ((bouton_VAL == 1) && (test_luminosite == 2)) {                    // Si l'on est en mode 'Réglage luminosité', on vient de valider ce réglage. OK + Reset KVB.
        test_luminosite = 0;
        EEPROM.update(0, luminosite);
        delay(50);
        softwareReset(WDTO_60MS);                                           // Reset logiciel.
      }
      if ((bouton_VAL == 1) && (roue_valid == 1)) {
        roue_valid = 0;
        voyant_VAL = 0; voyant_VALc = 0;                                    // Voyant touche [VAL] éteint.
// Avec cette validation, on peut aussi envoyer les chiffres des roues codeuses sur la liaison série, ici.
// Envoi 7 chiffres = roue_V_D  roue_V_U  roue_L_U  roue_D_C  roue_D_D  roue_D_U  bt_vomema (VO = 1, ME = 2 et MA = 3).
#ifdef FONCTION_ENV_ROUES                                                   // FONCTION : Permet d'envoyer les 7 chiffres des roues codeuses sur la liaison série, une fois validé par le bouton [VAL].
        Serial.write('<')         ; Serial.write('0'+roue_V_D); Serial.write('0'+roue_V_U); Serial.write('0'+roue_L_U);
        Serial.write('0'+roue_D_C); Serial.write('0'+roue_D_D); Serial.write('0'+roue_D_U); Serial.write('0'+bt_vomema); Serial.write('>');
#endif
      }
#ifdef FONCTION_VAL_VAL                                                     // FONCTION : Permet d'acquitter le clignotement de la touche [VAL].
      if (bouton_VAL == 1) { voyant_VAL = 0; voyant_VALc = 0; }
#endif
    }
// -----------------------------------------------
    if (bouton_MV != bouton_MV_old) {
      bouton_MV_old = bouton_MV;
      if (bouton_MV == 1) Serial.write('B'); else Serial.write('b');                       // Envoi le code de la touche [MV] sur la liaison série.
#ifdef FONCTION_VOY_MV_FC                                                                  // FONCTION : Allume les touches [MV] et [FC] en cas d'appui dessus.
      if ((bouton_MV == 1) && (test_luminosite == 0)) voyant_MV = voyant_MV^1;             // Action sur le voyant de la touche, à chaque appui.
      if (bouton_MV == 1){ if (voyant_MV == 1) Serial.write('H'); else Serial.write('h');} // Envoi un code suivant si le bouton est allumé ou éteint.
#endif
      if ((bouton_MV == 1) && (test_luminosite == 2) && (luminosite > 0)) {luminosite--; afficheurTM.setupDisplay(true, luminosite);} // Si l'on est en mode 'Réglage luminosité'
    }
// -----------------------------------------------
    if (bouton_FC != bouton_FC_old) {
      bouton_FC_old = bouton_FC;
      if (bouton_FC == 1) Serial.write('C'); else Serial.write('c');                       // Envoi le code de la touche [FC] sur la liaison série.
#ifdef FONCTION_VOY_MV_FC                                                                  // FONCTION : Allume les touches [MV] et [FC] en cas d'appui dessus.
      if ((bouton_FC == 1) && (test_luminosite == 0)) voyant_FC = voyant_FC^1;             // Action sur le voyant de la touche, à chaque appui.
      if (bouton_FC == 1){ if (voyant_FC == 1) Serial.write('I'); else Serial.write('i');} // Envoi un code suivant si le bouton est allumé ou éteint.
#endif
      if ((bouton_FC == 1) && (test_luminosite == 2) && (luminosite < 7)) {luminosite++; afficheurTM.setupDisplay(true, luminosite);} // Si l'on est en mode 'Réglage luminosité'
    }
// -----------------------------------------------
// Les actions du bouton [TEST] restent locales au KVB.
    if (bouton_TEST != bouton_TEST_old) {
      bouton_TEST_old = bouton_TEST;
      if (bouton_TEST == 1) Serial.write('D'); else Serial.write('d'); // Envoi le code de la touche [TEST] sur la liaison série.
#ifdef FONCTION_TEST                                                   // FONCTION : Permet de lancer la fonction test avec l'appui de la touche [TEST].
      if (bouton_TEST == 1) {
        test_local = 1; timer_test_local = courante_milli + duree_timer_test_local;
        analogWrite(SORTIE_HP, 30);                                    // Génération du signal PWM, pour alarme sonore. 30=ON.
        klaxon = 1; buzzer = 1;                                        // Active le HP et le buzzer via le MCP23017 + buzzer.
        afficheurTM.setDisplayToString("ss888888");                    // Allume tous les voyants et les lampes + affichage = [888][888].
        digitalWrite(SORTIE_LSSF, HIGH);                               // Voyant LSSF allumé.
      }
#endif
    }
// -----------------------------------------------
    if (bouton_SF != bouton_SF_old) {
      bouton_SF_old = bouton_SF;
      if (bouton_SF == 1) Serial.write('E'); else Serial.write('e');    // Envoi le code de la touche [SF] sur la liaison série.
#ifdef FONCTION_SF_LSSF                                                 // FONCTION : Un appui sur le bouton [SF] éteint le voyant LSSF.
      if (bouton_SF == 1) ampoule_LSSF = 0;
#endif
    }
  }

// -----------------------------------------------
  if (mcp23017_ok == 1) {                              // Si le boitier des roues codeuses est connecté, lecture des roues codeuses.
    lect_roues_cod();
#ifdef FONCTION_VAL_ROUES                              // FONCTION : Permet de faire clignoter le bouton [VAL] si l'on modifie les roues codeuses.
    i = roue_V_D + roue_V_U + roue_L_U + roue_D_C + roue_D_D + roue_D_U + bt_vomema;
    if (roue_cheksum != i) {
      roue_cheksum = i;                                // Une roue codeuse a bougé.
      roue_valid = 1;                                  // On doit appuyer sur [VAL] pour aquitter.
      voyant_VAL = 1; voyant_VALc = 1;                 // Voyant touche [VAL] Clignotant.
    }
#endif
  }
// -----------------------------------------------
  if (sequence_init_num != 0) {                        // Si 'sequence_init_num != 0', on est en séquence d'initilisation du KVB.
// La variable 'courante_milli' est mise à jour dans la routine = lecture_boutons(). Pas besoin de refaire appel ici à cette routine.
    if (courante_milli > timer_sequence_init) {
      sequence_init();
    }
  }
}

// ------------------------------------------------------------------------------------------------------------------------------------
// Sous-programme déclenché lors de la réception du caractère 's' = Séquence d'initialisation.
// La variable 'courante_milli' est mise à jour dans la routine = lecture_boutons(). Pas besoin de refaire appel ici à cette routine.
// sequence_init_num = n° de séquence en cours.
void sequence_init() {
  switch (sequence_init_num) {
    case 0 :
      lampe_V    = 0; lampe_FU    = 0; lampe_SOL = 0; lampe_ENG = 0;
      voyant_VAL = 0; voyant_VALc = 0; voyant_MV = 0; voyant_FC = 0;
      analogWrite(SORTIE_HP, 0);
      klaxon = 0; buzzer = 1;                             // Arrêt du HP et activation du le buzzer via le MCP23017.
      strncpy(buffer, "   PA200", 8); aff_clignotant = 0; // Affichage = [ PA][200].
      ampoule_LSSF = 2;                                   // Voyant LSSF clignotant.
      sequence_init_num++;
      timer_sequence_init = courante_milli + 1500 ;
      break;
    case 1 :
      strncpy(buffer, "        ", 8);                     // Eteint tout sur la façade.
      sequence_init_num++;
      timer_sequence_init = courante_milli + 200 ;
      break;
    case 2 :
      lampe_SOL = 1; lampe_ENG = 1;                       // Allume les lampes (SOL) + (ENG).
      strncpy(buffer, "   UC512", 8);                     // Affichage = [ UC][512].
      klaxon = 1;                                         // Activation du HP.
      sequence_init_num++;
      timer_sequence_init = courante_milli + 1500 ;
      break;
    case 3 :
      lampe_SOL = 0; lampe_ENG = 0;                       // Eteint les lampes (SOL) + (ENG).
      strncpy(buffer, "        ", 8);                     // Affichage = [   ][   ].
      klaxon = 0; buzzer = 0;                             // Arrêt du HP et du buzzer.
      voyant_VAL = 1; voyant_MV = 1; voyant_FC = 1;       // Allume birèvement les trois touches.
      sequence_init_num++;
      timer_sequence_init = courante_milli + 1000 ;
      break;
    case 4 :
      voyant_VALc = 1; voyant_MV = 0; voyant_FC = 0;      // En fin d'initialisation, la touche [VAL] et le voyant LSSF clignotent.
      sequence_init_num = 0;
      break;
    default:
      break;
  }
}

// ------------------------------------------------------------------------------------------------------------------------------------
// Sous-programme de lecture des boutons en face avant du KVB, à partir du circuit TM1638.
// On passe ici le plus souvent possible.
// A l'appui d'une touche la variable correspondante passe à '1'.
// Au relachement d'une touche la variable correspondante repasse à '0', après x msec. x = Variable "duree_antirebond" modifiable = 300 msec par défaut.

void lecture_boutons() {
  if (tm1638_ok == 1) etat_boutons = afficheurTM.getButtons();  // etat_boutons => 8 bits '0/1', '1' = Bouton enfoncé.
  courante_milli = millis();
// Traitement anti rebonds, avec un timer par bouton poussoir.
// Il faut au moins la durée 'duree_antirebond' en milli secondes, pour que la touche soit condidérée comme relâchée;
// Si le bouton était déjà enfoncé, au bout de la temporisation, on remet le bouton_xx = 0. On le remet à 1 si la touche est enfoncée.
  if ((bouton_VAL == 1)  && (courante_milli - duree_antirebond > bouton_VAL_milli))  bouton_VAL  = 0;
  if ((bouton_MV == 1)   && (courante_milli - duree_antirebond > bouton_MV_milli))   bouton_MV   = 0;
  if ((bouton_FC == 1)   && (courante_milli - duree_antirebond > bouton_FC_milli))   bouton_FC   = 0;
  if ((bouton_TEST == 1) && (courante_milli - duree_antirebond > bouton_TEST_milli)) bouton_TEST = 0;
  if ((bouton_SF == 1)   && (courante_milli - duree_antirebond > bouton_SF_milli))   bouton_SF   = 0;
// Si le bouton est enfoncé, bouton_xx=1 et le retour à 0 ne peut se faire qu'au bout de la temporisation (300 msec par défaut).
  if (bitRead(etat_boutons, 0)) {bouton_VAL  = 1; bouton_VAL_milli  = courante_milli;}
  if (bitRead(etat_boutons, 1)) {bouton_MV   = 1; bouton_MV_milli   = courante_milli;}
  if (bitRead(etat_boutons, 2)) {bouton_FC   = 1; bouton_FC_milli   = courante_milli;}
  if (bitRead(etat_boutons, 3)) {bouton_TEST = 1; bouton_TEST_milli = courante_milli;}
  if (bitRead(etat_boutons, 4)) {bouton_SF   = 1; bouton_SF_milli   = courante_milli;}
}
// ------------------------------------------------------------------------------------------------------------------------------------
// Sous-programme, petite intro sympathique d'animation de l'afficheur, à la mise sous tension du KVB.
// Exigence de ce logiciel libre : Ne pas modifier cette routine svp.
void intro_jlf() {
  if (tm1638_ok == 1) {
    strncpy(buffer, "  JLF   ", 8);
    afficheurTM.setDisplayToString(buffer);
    delay(500);     // Attendre 500 milli secondes.
    strncpy(buffer_old, "   15  02  2025    ", 19);
    for (i = 0; i < 19; i++) {
      buffer[5] = buffer_old[i];
      buffer[6] = buffer_old[i+1];
      buffer[7] = buffer_old[i+2];
      afficheurTM.setDisplayToString(buffer);
      delay(180);  // Attendre 180 milli secondes.
    }
    delay(500);    // Attendre 500 milli secondes.
    strncpy(buffer, "        ", 8);
    afficheurTM.setDisplayToString(buffer);
  }
}
// ------------------------------------------------------------------------------------------------------------------------------------
// Sous-programme pour réaliser un reset logiciel.
// Une fois le watchdog expiré, un reset est demandé automatiquement.
void softwareReset(uint8_t prescaller) {
  wdt_enable(prescaller);
  while (1) {}
}
// ------------------------------------------------------------------------------------------------------------------------------------
// Sous-programme pour la lecture des roues codeuses du boitier externe, et activation du haut-parleur.
// Roues Cod. : [VD VU]    [VU]    [DC DD DU]    Centaines, Dizaines, Unités + bouton VO-ME-MA
// Le compteur 'mcp23017_mode' permet d'alterner à chaque passage, l'écriture vers un MCP23017 et la lecture depuis l'autre MCP23017.

// MCP23017_ENTREES  Adresse = 0x21
// A0 à A7 = V_D, V_U, L_U, D_C, D_D, D_U, VOMEMA, X
// B0 à B7 = X, X, X, X, X, X, X, X

// MCP23017_SORTIES  Adresse = 0x20
// A0 à A7 = 0, 1, 2, 3, 4, 5, 6, 7
// B0 à B7 = 8, 9, X, X, X, HP fixe, Buzzer, HP Modulé

// B5 = 13 --->   HP    Sortie signal continu 0/5 Volts : 5 Volts = En marche.
// B6 = 14 ---> Buzzer  Sortie signal continu 0/5 Volts : 0 Volt  = En marche.
// B7 = 15 --->   HP    Sortie modulée Fréquence 1 Khz.

// HP modulé = Sortie modulée, on peut y brancher directement le HP avec un résistance de 150 Ohms en série.
// HP fixe   = Sortie continue 0 ou 5 Volts. Il faut ajouter un circuit de type sirène pour y brancher le HP.
// -----------------------------------------------
void lect_roues_cod() {
  if (mcp23017_mode == 0) {              // On active une seule patte à la fois, du MCP_SORTIES à 0.
// roue_1 = 0xC3FF - roue_cpt1, car par défaut le HP et le buzzer sont éteints. Dans ce cas, les sorties sont au niveau haut, +5 Volts = à '1'.
    roue_1 = 0xC3FF - roue_cpt1;         // val = de 0000 0011 1111 1110 à 0000 0001 1111 1111. Le '0' se décale de droite à gauche, de la position 0 à 9.
    if ((klaxon == 1) && (bitRead(roue_cpt2, 0) == 0)) bitWrite(roue_1, 15, 0) ; // Haut-Parleur modulé en fréquence, à chaque rafraichissement.
    if (buzzer == 1) bitWrite(roue_1, 14, 0);                                    // 0 = Buzzer actif en continu.
    if (klaxon == 1) bitWrite(roue_1, 13, 1);                                    // 1 = Haut-parleur actif en continu, demande un circuit de modulation externe.
    roue_2 = roue_1;                     // Pour inverser les valeurs du port A et du port B, pour suivre le plan de câblage.
    roue_2 = roue_2>>8;                  // Sur l'entier de 16 bits, on échange les 8 bits du haut avec les 8 bits du bas, pour que les 10 bits se suivent.
    roue_1 = roue_1<<8;
    roue_1 = roue_2 + roue_1;
    MCP_SORTIES.write16(roue_1) ;        // Attention les 8 bits forts sont cadrés à droite.
    if (bitRead(roue_cpt1, 9) == 0) { roue_cpt1 = roue_cpt1 << 1; } else { roue_cpt1 = 1; }  // roue_cpt1 de 0000 0000 0000 0001 à 0000 0010 0000 0000.
    roue_cpt2++;
    if (roue_cpt2 == 10) roue_cpt2 = 0;  // roue_cpt1 et roue_cpt2 vont ensemble !
    // Serial.print(val,BIN); Serial.print(" "); Serial.println(roue_cpt2);
  }
  if (mcp23017_mode == 1) {              // On va lire les entrées du MCP_ENTREES.
    roue_1 = MCP_ENTREES.read8(0);       // On lit seulement le PORT(A).
    //Serial.print(val,BIN);Serial.println(" ");
    if (bitRead(roue_1, 0) == 0) roue_D_U  = roue_cpt2; // A0
    if (bitRead(roue_1, 1) == 0) roue_D_D  = roue_cpt2; // A1
    if (bitRead(roue_1, 2) == 0) roue_D_C  = roue_cpt2; // A2
    if (bitRead(roue_1, 3) == 0) roue_L_U  = roue_cpt2; // A3
    if (bitRead(roue_1, 4) == 0) roue_V_U  = roue_cpt2; // A4
    if (bitRead(roue_1, 5) == 0) roue_V_D  = roue_cpt2; // A5
    if (bitRead(roue_1, 6) == 0) bt_vomema = roue_cpt2; // A6 - VO = 1, ME = 2 et MA = 3. L'entrée A7 n'est pas utilisée.
    //Serial.print(" V: "); Serial.print(roue_V_D); Serial.print(" "); Serial.print(roue_V_U); Serial.print(" L: "); Serial.print(roue_L_U); Serial.print(" D: ");
    //Serial.print(roue_D_C); Serial.print(" "); Serial.print(roue_D_D); Serial.print(" "); Serial.print(roue_D_U); Serial.print(" vomema "); Serial.println(bt_vomema);
  }
// On passe alernativement de l'écriture dans le MCP_SORTIES (mcp23017_mode = 0),
// puis au passage suivant dans le code, à la lecture du MCP_ENTREES (mcp23017_mode = 1).
// Comme cela on limite le temps perdu aux échanges I2C et on laisse le temps de s'établir aux signaux d'entrées du MCP23017.
  mcp23017_mode = mcp23017_mode^1;
}
